"Implementation of the local execution engine."

from .execution_engine import ExecutionEngine
from ..worker import worker_process


class LocalExecutionEngine(ExecutionEngine):
    "Execution engine that runs jobs on the local machine."
    def __call__(self, timeout, pending_work_items, config):
        for work_item in pending_work_items:
            yield worker_process(work_item, timeout, config)



"""Operator-provider plugin for all core cosmic ray operators.
"""

import itertools

from . import (binary_operator_replacement,
               boolean_replacer,
               break_continue,
               comparison_operator_replacement,
               exception_replacer,
               number_replacer,
               remove_decorator,
               unary_operator_replacement,
               zero_iteration_loop)


_OPERATORS = {
    op.__name__: op
    for op
    in itertools.chain(
        binary_operator_replacement.operators(),
        comparison_operator_replacement.operators(),
        unary_operator_replacement.operators(),
        (boolean_replacer.AddNot,
         boolean_replacer.ReplaceTrueFalse,
         boolean_replacer.ReplaceAndWithOr,
         boolean_replacer.ReplaceOrWithAnd,
         break_continue.ReplaceBreakWithContinue,
         break_continue.ReplaceContinueWithBreak,
         exception_replacer.ExceptionReplacer,
         number_replacer.NumberReplacer,
         remove_decorator.RemoveDecorator,
         zero_iteration_loop.ZeroIterationLoop))
}


class OperatorProvider:
    """Provider for all of the core Cosmic Ray operators."""

    def __iter__(self):
        return iter(_OPERATORS)

    def __getitem__(self, name):
        return _OPERATORS[name]



# An interceptor is a callable that is called at the end of the "init" command.
#
# Interceptors are passed the initialized WorkDB, and they are able to do
# things like mark certain mutations as skipped (i.e. so that they are never
# performed).



"Implementation of operator base class."

from abc import ABCMeta, abstractmethod
import ast


class Operator(ast.NodeTransformer):
    """
    A base class for all mutation operators.

    This takes care of the basic book-keeping that all operators need
    to do. All operators *must* derive from this class since that's
    how we keep track of them.
    """

    def __init__(self, core):
        self._core = core

    @property
    def core(self):
        """The core behavior of the operator."""
        return self._core

    def repr_args(self):
        "Extra arguments to display in operator reprs."
        return self.core.repr_args()

    def visit_mutation_site(self, node, num_mutations=1):
        """Subclasses call this when they encounter a node they can
         potentially mutate.

        This functions delegates to the core, letting it do whatever it needs
        to do.
        """
        return self.core.visit_mutation_site(node, self, num_mutations)

    def mutate(self, node, idx):
        """Mutate a node in an operator-specific manner.

        Return the new, mutated node. Return `None` if the node has
        been deleted. Return `node` if there is no mutation at all for
        some reason.
        """
        raise NotImplementedError(
            'Mutation operators must implement "mutate()".')

    def __repr__(self):
        repr_args = [('core', self.core.__class__.__name__)]
        repr_args.extend(self.core.repr_args())
        args = ['{}={}'.format(k, v)
                for k, v
                in repr_args]

        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join(args))


def _op_name(from_op, to_op):
    assert from_op or to_op, 'Cannot make replacement operator from None to None'

    if from_op is None:
        return 'Insert_{}'.format(to_op.__name__)
    elif to_op is None:
        return 'Delete_{}'.format(from_op.__name__)

    return '{}_{}'.format(from_op.__name__, to_op.__name__)


class ReplacementOperatorMeta(type):
    """Metaclass for mutation operators that replace Python operators.

    This does a few things:

    - Sets the name of the class object based on the class declaration *and* the from-/to-operators.
    - Makes the from-/to-operators available as class members.
    - Adds `Operator` as a base class.
    """
    def __init__(cls, name, bases, namespace, from_op, to_op, **kwargs):
        super().__init__(name, bases, namespace, **kwargs)

    def __new__(cls, name, bases, namespace, from_op, to_op, **kwargs):
        name = '{}_{}'.format(
            name,
            _op_name(from_op, to_op))
        bases = bases + (Operator,)
        namespace['from_op'] = from_op
        namespace['to_op'] = to_op
        return super().__new__(cls, name, bases, namespace, **kwargs)



"""Implementation of the binary-operator-replacement operator.
"""

import ast
import sys

from .operator import ReplacementOperatorMeta
from ..util import build_mutations

_AST_OPERATORS = (ast.Add, ast.Sub, ast.Mult, ast.Div, ast.FloorDiv, ast.Mod,
                  ast.Pow, ast.LShift, ast.RShift, ast.BitOr, ast.BitXor,
                  ast.BitAnd)

# todo: this often leads to unsupported syntax
if sys.version_info >= (3, 5):
    _AST_OPERATORS = _AST_OPERATORS + (ast.MatMult,)


def _to_ops(from_op):  # pylint: disable=unused-argument
    """The sequence of operators which `from_op` could be mutated to."""
    for to_op in _AST_OPERATORS:
        yield to_op


def _create_replace_binary_operator(from_op, to_op):
    class ReplaceBinaryOperator(
            metaclass=ReplacementOperatorMeta,
            from_op=from_op,
            to_op=to_op):
        """An operator that replaces binary operators."""
        def visit_BinOp(self, node):  # pylint: disable=invalid-name, missing-docstring
            if isinstance(node.op, self.from_op):
                return self.visit_mutation_site(node)
            return node

        def mutate(self, node, _):
            node.op = self.to_op()
            return node

    return ReplaceBinaryOperator


_MUTATION_OPERATORS = tuple(
    _create_replace_binary_operator(_AST_OPERATORS[idx], to_op)
    for idx, to_op in build_mutations(_AST_OPERATORS, _to_ops))


for op in _MUTATION_OPERATORS:
    globals()[op.__name__] = op


def operators():
    "Iterable of all binary operator replacement mutation operators."
    return iter(_MUTATION_OPERATORS)



"""Implementation of the WorkDB."""

import contextlib
import os
from enum import Enum

# This db may well not scale very well. We need to be ready to switch it out
# for something quicker if not. But for now it's *very* convenient.
import tinydb

from .work_item import WorkItem


class WorkDB:
    """WorkDB is the database that keeps track of mutation testing work progress.

    Essentially, there's a row in the DB for each mutation that needs to be
    executed in some run. These initially start off with no results, and
    results are added as they're completed.
    """
    class Mode(Enum):
        "Modes in which a WorkDB may be opened."

        # Open existing files, creating if necessary
        create = 1

        # Open only existing files, failing if it doesn't exist
        open = 2

    def __init__(self, path, mode):
        """Open a DB in file `path` in mode `mode`.

        Args:
          path: The path to the DB file.
          mode: The mode in which to open the DB. See the `Mode` enum for
            details.

        Raises:
          FileNotFoundError: If `mode` is `Mode.open` and `path` does not
            exist.
        """
        if (mode == WorkDB.Mode.open) and (not os.path.exists(path)):
            raise FileNotFoundError(
                'Requested file {} not found'.format(path))

        self._path = path
        self._db = tinydb.TinyDB(path)

    def close(self):
        """Close the database."""
        self._db.close()

    @property
    def name(self):
        """A name for this database.

        Derived from the constructor arguments.
        """
        return self._path

    @property
    def _config(self):
        """The table of work parameters."""
        return self._db.table('config')

    @property
    def _work_items(self):
        """The table of work items."""
        return self._db.table('work-items')

    @property
    def _pending(self):
        table = self._work_items
        work_item = tinydb.Query()
        # This somewhat tortured invocation is intended to appease linters.
        # They *hate* seeing "x == None" which is the natural expression of
        # this query, so we use this custom test instead.
        pending = table.search(
            work_item.worker_outcome.test(
                lambda val: val is None))
        return pending

    def set_config(self, config, timeout):
        """Set (replace) the configuration for the session.

        Args:
          config: Configuration object
          timeout: The timeout for tests.
        """
        table = self._config
        table.purge()
        table.insert({
            'config': config,
            'timeout': timeout,
        })

    def get_config(self):
        """Get the work parameters (if set) for the session.

        Returns: a tuple of `(config, timeout)`.

        Raises:
          ValueError: If is no config set for the session.
        """
        table = self._config

        try:
            record = table.all()[0]
        except IndexError:
            raise ValueError('work-db has no config')

        return (record['config'],
                record['timeout'])

    def add_work_items(self, work_items):
        """Add a sequence of WorkItems.

        Args:
          work_items: An iterable of WorkItems.
        """
        self._work_items.insert_multiple(work_items)

    def clear_work_items(self):
        """Clear all work items from the session.

        This removes any associated results as well.
        """
        self._work_items.purge()

    @property
    def work_items(self):
        """The sequence of WorkItems in the session.

        This include both complete and incomplete items.

        Each work item is a dict with the keys `module-name`, `op-name`, and
        `occurrence`. Items with results will also have the keys `results-type`
        and `results-data`.
        """
        return (WorkItem(r) for r in self._work_items)

    @property
    def num_work_items(self):
        """The number of WorkItems."""
        return len(self._work_items)

    def update_work_item(self, work_item):
        """Updates an existing WorkItem by job_id.

        Args:
            work_item: A WorkItem representing the new state of a job.

        Raises:
            KeyError: If there is no existing record with the same job_id.
        """
        self._work_items.update(
            work_item,
            tinydb.Query().job_id == work_item.job_id)

    @property
    def pending_work_items(self):
        """The sequence of pending WorkItems in the session."""
        return (WorkItem(r) for r in self._pending)

    @property
    def num_pending_work_items(self):
        """The number of pending WorkItems in the session."""
        return len(self._pending)


@contextlib.contextmanager
def use_db(path, mode=WorkDB.Mode.create):
    """
    Open a DB in file `path` in mode `mode` as a context manager.

    On exiting the context the DB will be automatically closed.

    Args:
      path: The path to the DB file.
      mode: The mode in which to open the DB. See the `Mode` enum for
        details.

    Raises:
      FileNotFoundError: If `mode` is `Mode.open` and `path` does not
        exist.
    """
    database = WorkDB(path, mode)
    try:
        yield database
    except Exception:
        raise
    finally:
        database.close()



"""Implementation of the replace-break-with-continue and
replace-continue-with-break operators.
"""

import ast

from .operator import Operator


class ReplaceBreakWithContinue(Operator):
    "Operator which replaces 'break' with 'continue'."

    def visit_Break(self, node):  # pylint: disable=invalid-name
        "Visit a 'break' node."
        return self.visit_mutation_site(node)

    def mutate(self, node, _):
        """Replace a Break node with a Continue node."""
        assert isinstance(node, ast.Break)
        return ast.Continue()


class ReplaceContinueWithBreak(Operator):
    "Operator which replaces 'continue' with 'break'."

    def visit_Continue(self, node):  # pylint: disable=invalid-name
        "Visit a 'continue' node."
        return self.visit_mutation_site(node)

    def mutate(self, node, _):
        """Replace a Continue node with a Break node."""
        assert isinstance(node, ast.Continue)
        return ast.Break()



from enum import IntEnum


class ExitCode(IntEnum):
    """Exit status codes.

    These mimic those on many unixes (and provided by `os`) but makes them
    available on all platforms.
    """
    # successful termination
    OK = 0

    # command line usage error
    Usage = 64

    # data format error
    DataErr = 65

    # cannot open input
    NoInput = 66

    # addressee unknown
    NoUser = 67

    # host name unknown
    NoHost = 68

    # service unavailable
    Unavailable = 69

    # internal software error
    Software = 70

    # system error (e.g., can't fork)
    OSErr = 71

    # critical OS file missing
    OSFile = 72

    # can't create (user) output file
    CantCreate = 73

    # input/output error
    IOErr = 74

    # temp failure; user is invited to retry
    TempFail = 75

    # remote error in protocol
    Protocol = 76

    # permission denied
    NoPerm = 77

    # configuration error
    Config = 78



"""Configuration module."""
from contextlib import contextmanager
import logging
import sys

import yaml

log = logging.getLogger()


@contextmanager
def _config_stream(filename):
    """Given a configuration's filename, this returns a stream from which a configuration can be read.

    If `filename` is `None` or '-' then stream will be `sys.stdin`. Otherwise,
    it's the open file handle for the filename.
    """
    if filename is None or filename == '-':
        log.info('Reading config from stdin')
        yield sys.stdin
    else:
        with open(filename, mode='rt') as handle:
            log.info('Reading config from %r', filename)
            yield handle


def load_config(filename=None):
    """Load a configuration from a file or stdin.

    If `filename` is `None`, then configuration gets read from stdin.

    Returns: A configuration dict.

    Raises:
      ConfigError: If there is an error loading the config.
    """
    try:
        with _config_stream(filename) as handle:
            filename = handle.name
            return yaml.safe_load(handle)
    except (OSError, UnicodeDecodeError, yaml.parser.ParserError) as exc:
        raise ConfigError(
            'Error loading configuration from {}'.format(filename)) from exc


def serialize_config(config):
    """Convert a configuration dict into a string.

    This is complementary with `load_config`.
    """
    return yaml.dump(config)


def get_db_name(session_name):
    """Determines the filename for a session.

    Basically, if `session_name` ends in ".json" this return `session_name`
    unchanged. Otherwise it return `session_name` with ".json" added to the
    end.
    """
    if session_name.endswith('.json'):
        return session_name
    return '{}.json'.format(session_name)


class ConfigError(Exception):
    """Errors loading configs or with values in a config."""
    pass



"""Query and retrieve the various plugins in Cosmic Ray.
"""

import logging

from stevedore import driver, ExtensionManager

log = logging.getLogger()


def _log_extension_loading_failure(_mgr, ep, err):
    # We have to log at the `error` level here as opposed to, say, `info`
    # because logging isn't configure when we reach here. We need this infor to
    # print with the default logging settings.
    log.error('Operator provider load failure: extension-point="%s", err="%s"', ep, err)


OPERATOR_PROVIDERS = {
    extension.name: extension.plugin()
    for extension in
    ExtensionManager(
        'cosmic_ray.operator_providers',
        on_load_failure_callback=_log_extension_loading_failure)
}


def get_operator(name):
    """Get an operator class from a provider plugin.

    Attrs:
        name: The name of the operator class.

    Returns: The operator *class object* (i.e. not an instance).
    """
    sep = name.index('/')
    provider_name = name[:sep]
    operator_name = name[sep + 1:]

    provider = OPERATOR_PROVIDERS[provider_name]
    return provider[operator_name]


def operator_names():
    """Get an iterable of all operator names."""
    return (
        '{}/{}'.format(provider_name, operator_name)
        for provider_name, provider in OPERATOR_PROVIDERS.items()
        for operator_name in provider)


def get_test_runner(name, test_args):
    """Get a test-runner instance by name."""
    test_runner_manager = driver.DriverManager(
        namespace='cosmic_ray.test_runners',
        name=name,
        invoke_on_load=True,
        invoke_args=(test_args,),
        on_load_failure_callback=_log_extension_loading_failure,
    )

    return test_runner_manager.driver


def test_runner_names():
    """Get iterable of test-runner plugin names."""
    return ExtensionManager(
        'cosmic_ray.test_runners',
        on_load_failure_callback=_log_extension_loading_failure,
    ).names()


def get_interceptor(name):
    """Get an interceptor by name.

    Attrs:
        name: The name of the plugin containing the interceptor.

    Returns: A callable object which must accept a single `WorkDB` argument.
    """
    return ExtensionManager(
        'cosmic_ray.interceptors',
        on_load_failure_callback=_log_extension_loading_failure,
    )[name].plugin


def interceptor_names():
    """Get an iterable of all interceptor plugin names."""
    return ExtensionManager(
        'cosmic_ray.interceptors',
        on_load_failure_callback=_log_extension_loading_failure,
    ).names()


def get_execution_engine(name):
    """Get the execution engine by name."""
    manager = driver.DriverManager(
        namespace='cosmic_ray.execution_engines',
        name=name,
        invoke_on_load=True,
        on_load_failure_callback=_log_extension_loading_failure,
    )

    return manager.driver


def execution_engine_names():
    """Get iterable of execution-enginer plugin names."""
    return ExtensionManager(
        'cosmic_ray.execution_engines',
        on_load_failure_callback=_log_extension_loading_failure,
    ).names()



"""This module contains mutation operators which replace one
comparison operator with another.
"""

import ast

from ..util import build_mutations, compare_ast
from .operator import ReplacementOperatorMeta


_AST_OPERATORS = (ast.Eq, ast.NotEq, ast.Lt, ast.LtE, ast.Gt, ast.GtE,
                  ast.Is, ast.IsNot, ast.In, ast.NotIn)


def _all_ops(from_op):
    """The sequence of operators which `from_op` could be mutated to.

    There are a number of potential replacements which we avoid because they
    almost always produce equivalent mutants. We encode those in the logic
    here.

    NB: This is an imperfect, stop-gap solution to the problem of certain
    equivalent mutants. Obviously `==` is not generally the same as `is`, but
    that mutation is also a source of a good number of equivalents. The real
    solution to this problem will probably come in the form of real exception
    declarations or something.

    See https://github.com/sixty-north/cosmic-ray/pull/162 for some more
    discussion of this issue.

    """

    to_ops = set(to_op
                 for idx, to_op
                 in build_mutations([from_op],
                                    lambda _: _AST_OPERATORS))
    for to_op in to_ops:
        if from_op is ast.Eq and to_op is ast.Is:
            pass
        elif from_op is ast.NotEq and to_op is ast.IsNot:
            pass
        else:
            yield to_op


# Maps from-ops to to-ops when the RHS is `None`
_RHS_IS_NONE_OPS = {
    ast.Eq: [ast.IsNot],
    ast.NotEq: [ast.Is],
    ast.Is: [ast.IsNot],
    ast.IsNot: [ast.Is],
}

_RHS_IS_INTEGER_OPS = (ast.Eq, ast.NotEq, ast.Lt, ast.LtE, ast.Gt, ast.GtE)


def _rhs_is_none_ops(from_op):
    assert issubclass(from_op, ast.AST)
    return _RHS_IS_NONE_OPS.get(from_op, ())


def _rhs_is_integer_ops(from_op):
    assert issubclass(from_op, ast.AST)
    return _RHS_IS_INTEGER_OPS


def _comparison_rhs_is_none(node):
    "Determine if the node is a comparison with `None` on the RHS."
    return ((len(node.comparators) == 1)
            and
            (compare_ast(node.comparators[0], ast.NameConstant(None))))


def _comparison_rhs_is_integer(node):
    return ((len(node.comparators) == 1)
            and
            isinstance(node.comparators[0], ast.Num))


def _build_mutations(node):
    """Given a Compare node, produce the list of mutated operations.

    Depending on the details of the Compare node, different tactics
    may be used to generate the list of mutations, in order to avoid
    generating mutants which we know will be incompetent.

    Args:
        node: A Compare node.

    Returns:
        A sequence of (idx, to-op) tuples describing the mutations for `ops`.
        The idx is the index into the list of ops for the Compare node
        (A single Compare node can contain multiple operators in order to
        represent expressions like 5 <= x < 10).
    """
    assert isinstance(node, ast.Compare)
    ops = _find_to_ops(node)
    return build_mutations(map(type, node.ops), ops)


def _find_to_ops(node):
    """Iterable of possible operators the node could be mutated to.
    """
    if _comparison_rhs_is_none(node):
        ops = _rhs_is_none_ops
    elif _comparison_rhs_is_integer(node):
        ops = _rhs_is_integer_ops
    else:
        ops = _all_ops
    return ops


def _create_replace_comparison_operator(from_op, to_op):
    class ReplaceComparisonOperator(
            metaclass=ReplacementOperatorMeta,
            from_op=from_op,
            to_op=to_op):  # pylint: disable=abstract-method
        """An operator that modifies comparisons."""

        def _mutations(self, node):
            """List of all mutations for a node that this operator should perform.

            Returns a list of `(idx, to_op)` where `idx` is an index into
            `node.ops` and `to_op` is an operator node class.

            This limits mutations to those from `from_op` to `to_op`.
            """
            return [(idx, to_op)
                    for idx, to_op
                    in _build_mutations(node)
                    if isinstance(node.ops[idx], self.from_op)
                    if to_op is self.to_op]

        def visit_Compare(self, node):  # pylint: disable=missing-docstring
            muts = self._mutations(node)
            if muts:
                return self.visit_mutation_site(node, len(muts))

            return node

        def mutate(self, node, idx):
            op_idx, to_op = self._mutations(node)[idx]
            assert isinstance(node.ops[op_idx], self.from_op)
            node.ops[op_idx] = to_op()
            return node

    return ReplaceComparisonOperator


# Create the operator classes
_MUTATION_OPERATORS = tuple(
    _create_replace_comparison_operator(from_op, to_op)
    for from_op in _AST_OPERATORS
    for to_op in _all_ops(from_op))


# Add them to the module namespace
for op in _MUTATION_OPERATORS:
    globals()[op.__name__] = op


def operators():
    "Iterable of all comparison operator replacement mutation operators."
    return iter(_MUTATION_OPERATORS)



"""Implementation of the NumberReplacer operator.
"""

import ast

from .operator import Operator


# List of offsets that we apply to numbers in the AST. Each index into the list
# corresponds to single mutation.
OFFSETS = [
    +1,
    -1,
]


class NumberReplacer(Operator):
    """An operator that modifies numeric constants."""

    def visit_Num(self, node):  # noqa # pylint: disable=invalid-name
        "Visit a number node."
        return self.visit_mutation_site(node, len(OFFSETS))

    def mutate(self, node, idx):
        """Modify the numeric value on `node`."""

        assert idx < len(OFFSETS), 'received count with no associated offset'

        offset = OFFSETS[idx]
        new_node = ast.Num(n=node.n + offset)
        return new_node



"""This is the command-line program for cosmic ray.

Here we manage command-line parsing and launching of the internal
machinery that does mutation testing.
"""
import itertools
import json
import logging
import os
import pprint
import signal
import subprocess
import sys

import docopt
import docopt_subcommands as dsc

import cosmic_ray.commands
import cosmic_ray.counting
import cosmic_ray.modules
import cosmic_ray.plugins
import cosmic_ray.worker
from cosmic_ray.config import ConfigError, get_db_name, load_config
from cosmic_ray.exit_codes import ExitCode
from cosmic_ray.progress import report_progress
from cosmic_ray.testing.test_runner import TestOutcome
from cosmic_ray.timing import Timer
from cosmic_ray.util import redirect_stdout
from cosmic_ray.work_db import use_db, WorkDB
from cosmic_ray.version import __version__

log = logging.getLogger()


@dsc.command()
def handle_baseline(args):
    """usage: cosmic-ray baseline <config-file>

    Run an un-mutated baseline of the specific configuration. This is
    largely like running a "worker" process, with the difference that
    a baseline run doesn't mutate the code.

    """
    sys.path.insert(0, '')

    config = load_config(args['<config-file>'])

    test_runner = cosmic_ray.plugins.get_test_runner(
        config['test-runner']['name'],
        config['test-runner']['args'])
    work_item = test_runner()
    # note: test_runner() results are meant to represent
    # status codes when executed against mutants.
    # SURVIVED means that the test suite executed without any error
    # hence CR thinks the mutant survived. However when running the
    # baseline execution we don't have mutations and really want the
    # test suite to report PASS, hence the comparison below!
    if work_item.test_outcome != TestOutcome.SURVIVED:
        # baseline failed, print whatever was returned
        # from the test runner and exit
        log.error('baseline failed')
        print(''.join(work_item.data))
        return 2

    return ExitCode.OK


@dsc.command()
def handle_new_config(args):
    """usage: cosmic-ray new-config <config-file>

    Create a new config file.
    """
    config = cosmic_ray.commands.new_config()
    with open(args['<config-file>'], mode='wt') as handle:
        handle.write(config)

    return ExitCode.OK


@dsc.command()
def handle_init(args):
    """usage: cosmic-ray init <config-file> <session-file>

    Initialize a mutation testing session from a configuration. This
    primarily creates a session - a database of "work to be done" -
    which describes all of the mutations and test runs that need to be
    executed for a full mutation testing run. The configuration
    specifies the top-level module to mutate, the tests to run, and how
    to run them.

    This command doesn't actually run any tests. Instead, it scans the
    modules-under-test and simply generates the work order which can be
    executed with other commands.

    The `session-file` is the filename for the database in which the
    work order will be stored.
    """
    # This lets us import modules from the current directory. Should
    # probably be optional, and needs to also be applied to workers!
    sys.path.insert(0, '')

    config_file = args['<config-file>']

    config = load_config(config_file)

    if 'timeout' in config:
        timeout = float(config['timeout'])
    elif 'baseline' in config:
        try:
            baseline_mult = float(config['baseline'])
            if baseline_mult <= 0:
                raise ValueError()
        except (ValueError, TypeError):
            raise ConfigError(
                'Baseline multiplier must be a positive number, not {}'.format(
                    config['baseline']))

        command = 'cosmic-ray baseline {}'.format(
            args['<config-file>'])

        # We run the baseline in a subprocess to more closely emulate the
        # runtime of a worker subprocess.
        with Timer() as timer:
            subprocess.check_call(command.split())

        timeout = baseline_mult * timer.elapsed.total_seconds()
    else:
        raise ConfigError(
            "Config must specify either baseline or timeout")

    log.info('timeout = %f seconds', timeout)

    modules = set(
        cosmic_ray.modules.find_modules(
            cosmic_ray.modules.fixup_module_name(config['module']),
            config.get('exclude-modules', None)))

    log.info('Modules discovered: %s', [m.__name__ for m in modules])

    db_name = get_db_name(args['<session-file>'])

    with use_db(db_name) as database:
        cosmic_ray.commands.init(
            modules,
            database,
            config,
            timeout)

    return ExitCode.OK


@dsc.command()
def handle_config(args):
    """usage: cosmic-ray config <session-file>

    Show the configuration for in a session.
    """
    session_file = get_db_name(args['<session-file>'])
    with use_db(session_file) as database:
        config, _ = database.get_config()
        print(json.dumps(config))

    return ExitCode.OK


@dsc.command()
def handle_exec(args):
    """usage: cosmic-ray exec <session-file>

    Perform the remaining work to be done in the specified session.
    This requires that the rest of your mutation testing
    infrastructure (e.g. worker processes) are already running.
    """

    session_file = get_db_name(
        args.get('<session-file>'))
    cosmic_ray.commands.execute(session_file)

    return ExitCode.OK


@dsc.command()
def handle_dump(args):
    """usage: cosmic-ray dump <session-file>

    JSON dump of session data. This output is typically run through
    other programs to produce reports.
    """
    session_file = get_db_name(args['<session-file>'])

    with use_db(session_file, WorkDB.Mode.open) as database:
        for record in database.work_items:
            print(json.dumps(record))

    return ExitCode.OK


@dsc.command()
def handle_counts(args):
    """usage: {program} counts <config-file>

    Count the number of tests that would be run for a given testing
    configuration. This is mostly useful for estimating run times and
    keeping track of testing statistics.
    """
    config = load_config(args['<config-file>'])

    sys.path.insert(0, '')

    modules = cosmic_ray.modules.find_modules(
        cosmic_ray.modules.fixup_module_name(config['module']),
        config.get('exclude-modules', []))

    operators = cosmic_ray.plugins.operator_names()

    counts = cosmic_ray.counting.count_mutants(modules, operators)

    print('[Counts]')
    pprint.pprint(counts)
    print('\n[Total test runs]\n',
          sum(itertools.chain(
              *(d.values() for d in counts.values()))))

    return ExitCode.OK


@dsc.command()
def handle_test_runners(args):
    """usage: {program} test-runners

    List the available test-runner plugins.
    """
    assert args
    print('\n'.join(cosmic_ray.plugins.test_runner_names()))

    return ExitCode.OK


@dsc.command()
def handle_operators(args):
    """usage: {program} operators

    List the available operator plugins.
    """
    assert args
    print('\n'.join(cosmic_ray.plugins.operator_names()))

    return ExitCode.OK


@dsc.command()
def handle_execution_engines(args):
    """usage: {program} execution-engines

    List the available execution-engine plugins.
    """
    assert args
    print('\n'.join(cosmic_ray.plugins.execution_engine_names()))

    return ExitCode.OK


@dsc.command()
def handle_interceptors(args):
    """usage: {program} interceptors

    List the available interceptor plugins.
    """
    assert args
    print('\n'.join(cosmic_ray.plugins.interceptor_names()))

    return ExitCode.OK


@dsc.command()
def handle_worker(args):
    """usage: {program} worker \
    [options] <module> <operator> <occurrence> [<config-file>]

    Run a worker process which performs a single mutation and test run.
    Each worker does a minimal, isolated chunk of work: it mutates the
    <occurence>-th instance of <operator> in <module>, runs the test
    suite defined in the configuration, prints the results, and exits.

    Normally you won't run this directly. Rather, it will be launched
    by an execution engine. However, it can be useful to run this on
    its own for testing and debugging purposes.

    options:
      --keep-stdout       Do not squelch stdout

    """
    config = load_config(args.get('<config-file>'))

    if config.get('local-imports', True):
        sys.path.insert(0, '')

    operator = cosmic_ray.plugins.get_operator(args['<operator>'])
    test_runner = cosmic_ray.plugins.get_test_runner(
        config['test-runner']['name'],
        config['test-runner']['args'])

    with open(os.devnull, 'w') as devnull:
        with redirect_stdout(sys.stdout if args['--keep-stdout'] else devnull):
            work_item = cosmic_ray.worker.worker(
                args['<module>'],
                operator,
                int(args['<occurrence>']),
                test_runner)

    sys.stdout.write(json.dumps(work_item))

    return ExitCode.OK


DOC_TEMPLATE = """{program}

Usage: {program} [options] <command> [<args> ...]

Options:
  -h --help     Show this screen.
  -v --verbose  Use verbose logging

Available commands:
  {available_commands}

See '{program} help <command>' for help on specific commands.
"""


def common_option_handler(config):
    """Add verbose mode.

    :param config: holds the configuration values
    """
    if config['--verbose']:
        logging.basicConfig(level=logging.INFO)


_SIGNAL_EXIT_CODE_BASE = 128


def main(argv=None):
    """ Invoke the cosmic ray evaluation.

    :param argv: the command line arguments
    """
    signal.signal(signal.SIGINT,
                  lambda *args: sys.exit(_SIGNAL_EXIT_CODE_BASE + signal.SIGINT))

    if hasattr(signal, 'SIGINFO'):
        signal.signal(getattr(signal, 'SIGINFO'),
                      lambda *args: report_progress(sys.stderr))

    try:
        return dsc.main(
            'cosmic-ray',
            'Cosmic Ray {}'.format(__version__),
            argv=argv,
            doc_template=DOC_TEMPLATE,
            common_option_handler=common_option_handler,
            exit_at_end=False)
    except docopt.DocoptExit as exc:
        print(exc, file=sys.stderr)
        return ExitCode.Usage
    except FileNotFoundError as exc:
        print(exc, file=sys.stderr)
        return ExitCode.NoInput
    except PermissionError as exc:
        print(exc, file=sys.stderr)
        return ExitCode.NoPerm
    except ConfigError as exc:
        print(exc, file=sys.stderr)
        if exc.__cause__ is not None:
            print(exc.__cause__, file=sys.stderr)
        return ExitCode.Config


if __name__ == '__main__':
    sys.exit(main())



"""Various utility functions with no better place to live.
"""
import ast
import itertools

try:
    from contextlib import redirect_stdout
except ImportError:

    import sys


    # noqa # redirect_stdout was introduced in Python 3.4
    class _RedirectStream:
        """
            Copied from Python 3.5's implementation. See:
            https://github.com/python/cpython/commit/83935e76e35cf8d2fb9fe2599420f8adf421b884#diff-edbcdd20abc32f8b018deb2353ae925a
        """
        _stream = None

        def __init__(self, new_target):
            self._new_target = new_target
            # We use a list of old targets to make this CM re-entrant
            self._old_targets = []

        def __enter__(self):
            self._old_targets.append(getattr(sys, self._stream))
            setattr(sys, self._stream, self._new_target)
            return self._new_target

        def __exit__(self, exctype, excinst, exctb):
            setattr(sys, self._stream, self._old_targets.pop())


    # noqa # pylint: disable=invalid-name
    class redirect_stdout(_RedirectStream):  # noqa
        """Context manager for temporarily redirecting stdout to another
        file."""
        _stream = "stdout"

try:
    # pylint: disable=unused-import, ungrouped-imports
    from contextlib import redirect_stderr
except ImportError:
    # redirect_stderr was introduced in Python 3.5
    # pylint: disable=invalid-name
    class redirect_stderr(redirect_stdout):
        """
            Copied from Python 3.5's implementation. See:
            https://github.com/python/cpython/commit/83935e76e35cf8d2fb9fe2599420f8adf421b884#diff-edbcdd20abc32f8b018deb2353ae925a
        """

        _stream = "stderr"


def get_line_number(node):
    """Try to get the line number for `node`.

    If no line number is available, this returns -1.
    """
    try:
        return node.lineno
    except AttributeError:
        return -1


def get_col_offset(node):
    """Try to get the column offset for `node`.

    If no column offset is available, this returns -1.
    """
    try:
        return node.col_offset
    except AttributeError:
        return -1


def build_mutations(ops, to_ops):
    """The sequence of `(idx, to-op)` tuples describing the mutations for `ops`.

    Each `idx` is an index into `ops` indicating the operator to be mutated
    from. Each `to-op` is the operator class to be mutated to.

    This will never produce mutations from an operator to itself.

    If `to_ops` includes `None` in its output, that means to delete the
    from-operator.

    Args:
      ops: A sequence of AST operator classes
      to_ops: A unary callable which takes a from-operator class and returns an
        iterable of operator classes (or `None`) to mutate to.
    """
    return [(idx, to_op)
            for idx, from_op in enumerate(ops)
            for to_op in to_ops(from_op)
            if from_op is not to_op]


def compare_ast(node1, node2):
    """Compares two AST nodes for equality.

    Ignores the lineno, col_offset and ctx attributes.

    Args:
        node1: An AST node.
        node2: Another AST node.

    Returns:
        True if the nodes are equivalent, otherwise False.
    """
    if type(node1) is not type(node2):
        return False
    if isinstance(node1, ast.AST):
        for k, v in vars(node1).items():
            if k in ('lineno', 'col_offset', 'ctx'):
                continue
            if not compare_ast(v, getattr(node2, k)):
                return False
        return True
    elif isinstance(node1, list):
        return all(itertools.starmap(compare_ast, zip(node1, node2)))
    return node1 == node2



"Implementation of the 'execute' command."
import os

from cosmic_ray.progress import reports_progress
from cosmic_ray.work_db import use_db, WorkDB
from cosmic_ray.plugins import get_execution_engine


_progress_messages = {}  # pylint: disable=invalid-name


def _update_progress(work_db):
    pending = work_db.num_pending_work_items
    total = work_db.num_work_items
    remaining = total - pending
    message = "{} out of {} completed".format(remaining, total)
    _progress_messages[work_db.name] = message


def _report_progress(stream):
    for db_name, progress_message in _progress_messages.items():
        session = os.path.splitext(db_name)[0]
        print("{session} : {progress_message}"
              .format(session=session,
                      progress_message=progress_message),
              file=stream)


@reports_progress(_report_progress)
def execute(db_name):
    """Execute any pending work in the database stored in `db_name`,
    recording the results.

    This looks for any work in `db_name` which has no results, schedules it to
    be executed, and records any results that arrive.
    """
    try:
        with use_db(db_name, mode=WorkDB.Mode.open) as work_db:
            _update_progress(work_db)
            config, timeout = work_db.get_config()
            engine_config = config['execution-engine']
            executor = get_execution_engine(engine_config['name'])
            work_items = executor(timeout,
                                  work_db.pending_work_items,
                                  config)

            for work_item in work_items:
                work_db.update_work_item(work_item)
                _update_progress(work_db)
    except FileNotFoundError as exc:
        raise FileNotFoundError(str(exc).replace(
            'Requested file', 'Corresponding database', 1)) from exc



"""A NodeVisitor that prints information.

Primarily useful for debugging and diagnostics.
"""

import ast


class PrintVisitor(ast.NodeVisitor):
    """A `NodeVisitor` which simply prints information about the nodes it
    visits.
    """

    def __init__(self):
        self.indent = ''

    def generic_visit(self, node):
        """Print `node` and manage an indentation state."""
        print(self.indent, repr(node))
        self.indent += '    '
        super().generic_visit(node)  # pylint:disable=missing-super-argument
        self.indent = self.indent[:-4]

    def visit_Num(self,  # noqa # pylint:disable=invalid-name,no-self-use
                  node):
        "Special print handling for numbers."
        print('a number:', node)


def dump_mod():
    """Print an entire module with the PrintVisitor.
    """
    import mod  # pylint:disable=import-error
    with open(mod.__file__, 'rt') as module_file:
        nodes = ast.parse(module_file.read())
    PrintVisitor().visit(nodes)



"""Implementation of the unary-operator-replacement operator.
"""

import ast

from ..util import build_mutations
from .operator import ReplacementOperatorMeta


# None indicates we want to delete the operator
_AST_OPERATORS = (ast.UAdd, ast.USub, ast.Invert, ast.Not)


def _to_ops(from_op):
    """The sequence of operators which `from_op` could be mutated to.
    """

    for to_op in _AST_OPERATORS + (None,):
        if to_op and from_op is ast.Not:
            # 'not' can only be removed but not replaced with
            # '+', '-' or '~' b/c that may lead to strange results
            pass
        elif from_op is ast.UAdd and to_op is None:
            # '+1' => '1' yields equivalent mutations
            pass
        else:
            yield to_op


def _create_replace_unary_operator(from_op, to_op):
    """Create a ReplaceUnaryOperator subclasses that mutates `from_op` to `to_op`.
    """
    class ReplaceUnaryOperator(
            metaclass=ReplacementOperatorMeta,
            from_op=from_op,
            to_op=to_op):  # pylint: disable=abstract-method
        """An operator that modifies unary operators.
        """

        def visit_UnaryOp(self, node):  # pylint: disable=invalid-name, missing-docstring
            if isinstance(node.op, self.from_op):
                return self.visit_mutation_site(node)
            return node

        def mutate(self, node, _):
            "Perform the `idx`th mutation on node."
            if self.to_op is None:
                return node.operand

            node.op = self.to_op()
            return node

    return ReplaceUnaryOperator


# Create the mutation operator classes
_MUTATION_OPERATORS = tuple(
    _create_replace_unary_operator(_AST_OPERATORS[idx], to_op)
    for idx, to_op in build_mutations(_AST_OPERATORS, _to_ops))


# Add the mutation operator classes to the module namespace.
for op in _MUTATION_OPERATORS:
    globals()[op.__name__] = op


def operators():
    "Iterable of all unary operator replacement mutation operators."
    return iter(_MUTATION_OPERATORS)



"Implementation of the 'init' command."
import logging
import uuid

import cosmic_ray.modules
from cosmic_ray.parsing import get_ast
from cosmic_ray.plugins import get_interceptor, interceptor_names, get_operator
from cosmic_ray.util import get_col_offset, get_line_number
from cosmic_ray.work_item import WorkItem


log = logging.getLogger()


class WorkDBInitCore:
    """Operator core that initializes a WorkDB for a specific module and operator.

    The idea is to walk the AST looking for nodes that the operator can mutate.
    As they're found, `visit_mutation_site` is called and this core adds new
    WorkItems to the WorkDB. Use this core to populate a WorkDB by creating one
    for each operator-module pair and running it over the module's AST.
    """
    def __init__(self, module, op_name, work_db):
        self.module = module
        self.op_name = op_name
        self.work_db = work_db
        self.occurrence = 0

    def visit_mutation_site(self, node, _, count):
        """Adds work items to the WorkDB as mutatable nodes are found.
        """
        self.work_db.add_work_items(
            WorkItem(
                job_id=uuid.uuid4().hex,
                module=self.module.__name__,
                operator=self.op_name,
                occurrence=self.occurrence + c,
                filename=self.module.__file__,
                line_number=get_line_number(node),
                col_offset=get_col_offset(node))
            for c in range(count))

        self.occurrence += count

        return node


def init(modules,
         work_db,
         config,
         timeout):
    """Clear and initialize a work-db with work items.

    Any existing data in the work-db will be cleared and replaced with entirely
    new work orders. In particular, this means that any results in the db are
    removed.

    Args:
      modules: iterable of module objects to be mutated.
      work_db: A `WorkDB` instance into which the work orders will be saved.
      config: The configuration for the new session.
      timeout: The timeout to apply to the work in the session.
    """
    operators = cosmic_ray.plugins.operator_names()
    work_db.set_config(
        config=config,
        timeout=timeout)

    work_db.clear_work_items()

    for module in modules:
        for op_name in operators:
            core = WorkDBInitCore(module, op_name, work_db)
            operator = get_operator(op_name)(core)
            module_ast = get_ast(module)
            operator.visit(module_ast)

    apply_interceptors(work_db)


def apply_interceptors(work_db):
    """Apply each registered interceptor to the WorkDB."""
    for name in interceptor_names():
        interceptor = get_interceptor(name)
        interceptor(work_db)



"""An interceptor that uses spor metadata to determine when specific mutations
should be skipped.
"""
import logging

from spor.repo import find_anchors

from cosmic_ray.worker import WorkerOutcome

log = logging.getLogger()


def intercept(work_db):
    """Look for WorkItems in `work_db` that should not be mutated due to spor metadata.

    For each WorkItem, find anchors for the item's file/line/columns. If an
    anchor exists with metadata containing `{mutate: False}` then the WorkItem
    is marked as SKIPPED.
    """
    for item in work_db.work_items:
        try:
            anchors = tuple(find_anchors(item.filename))
        except ValueError:
            log.info('No spor repository for %s', item.filename)
            continue

        for anchor in anchors:
            metadata = anchor.metadata
            if item.line_number == anchor.line_number and not metadata.get('mutate', True):
                item.worker_outcome = WorkerOutcome.SKIPPED
                log.info('skipping %s', item)
                work_db.update_work_item(item)



"Base execution-engine implementation details."

import abc


class ExecutionEngine(metaclass=abc.ABCMeta):
    "Base class for execution engine plugins."
    @abc.abstractmethod
    def __call__(self, timeout, pending_work, config):
        """Execute jobs in `pending_work_items`.

        Spend no more than `timeout` seconds for
        a single job, using `config` to control the work.
        """
        pass
